@@ -1,8 +1,14 @@
 
 import os
+import sys
+import time
 import logging
+import tempfile
 
-import orjson  # faster json serialization
+try:
+    import orjson  # faster json serialization
+except ImportError:
+    import json as orjson
 
 from py4j.java_gateway import launch_gateway
 from py4j.java_gateway import JavaGateway, GatewayParameters, CallbackServerParameters
@@ -25,7 +31,6 @@ class TextWorldExpressEnv:
         # Launch Java side with dynamic port and get back the port on which the
         # server was bound to.
         if DEBUG_MODE:
-            import sys, time
             port = launch_gateway(
                 classpath=serverPath, die_on_exit=True, cwd=BASEPATH,
                 javaopts=['-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=5005,quiet=y'],
@@ -33,13 +38,14 @@ class TextWorldExpressEnv:
             print("Attach debugger within the next 10 seconds")
             time.sleep(10)  # Give time for user to attach debugger
         else:
-            port = launch_gateway(classpath=serverPath, die_on_exit=True, cwd=BASEPATH)
+            port, proc = launch_gateway(classpath=serverPath, die_on_exit=True, cwd=BASEPATH, javaopts=['-Xverify:none'], return_proc=True)
 
         # Connect python side to Java side with Java dynamic port and start python
         # callback server with a dynamic port
         self._gateway = JavaGateway(
             gateway_parameters=GatewayParameters(auto_field=True, port=port),
-            callback_server_parameters=CallbackServerParameters(port=0, daemonize=True))
+            callback_server_parameters=CallbackServerParameters(port=0, daemonize=True),
+            java_process=proc)
 
         # Retrieve the port on which the python callback server was bound to.
         python_port = self._gateway.get_callback_server().get_listening_port()
@@ -56,9 +62,6 @@ class TextWorldExpressEnv:
         # Keep track of the last step score, to calculate reward from score
         self.lastStepScore = 0
 
-        # Load the script
-        #self.load(self.gameName, 0, "")
-
         # Set the environment step limit
         self.envStepLimit = envStepLimit
 
@@ -73,9 +76,11 @@ class TextWorldExpressEnv:
         self.seed = None
         self.gameName = None
         self.gameParams = ""
-        self.gameFold = None
+        self.gameFold = "train"
         self.generateGoldPath = False
 
+        self._obj_tree_tempfile = tempfile.NamedTemporaryFile()
+
     #
     #   Run History
     #
@@ -98,7 +103,7 @@ class TextWorldExpressEnv:
         return packed
 
     def addStepToHistory(self, step):
-        self.runHistory.append(step)
+        self.runHistory.append(dict(step))
 
     def getNumSteps(self):
         return len(self.runHistory)
@@ -140,6 +145,18 @@ class TextWorldExpressEnv:
         if msg:
             raise ValueError(msg)
 
+    def close(self) -> None:
+        self._gateway.shutdown()
+
+        # According to https://github.com/py4j/py4j/issues/320#issuecomment-553599210
+        # we need to send a newline to the process to make it exit.
+        if self._gateway.java_process.poll() is None:
+            self._gateway.java_process.stdin.write("\n".encode("utf-8"))
+            self._gateway.java_process.stdin.flush()
+
+    def __del__(self):
+        self.close()
+
 
     # Get a list of valid tasks/environments
     def getGameNames(self):
@@ -153,6 +170,16 @@ class TextWorldExpressEnv:
     def getTaskDescription(self):
         return self.server.getTaskDescription()
 
+    def getObjectTree(self):
+        msg = self.server.getObjectTree(self._obj_tree_tempfile.name)
+        if msg:
+            # Game is not initialized.
+            raise RuntimeError(msg)
+
+        self._obj_tree_tempfile.file.seek(0)
+        payload = self._obj_tree_tempfile.file.read()
+        return orjson.loads(payload)
+
     #
     # Train/development/test sets
     #
@@ -204,6 +231,12 @@ class TextWorldExpressEnv:
     # Step
     #
     def step(self, inputStr:str):
+        # Step 0: If `help` command, print the task description.
+        if inputStr == "help":
+            observation = self.getTaskDescription()
+            infos = dict(self.runHistory[-1])
+            return observation, 0, False, infos
+
         # Step 1: Take a step in the environment
         infos = self.parseJSONResponse(self.server.stepJSON(inputStr))
         observation = infos["observation"]
@@ -239,3 +272,33 @@ class TextWorldExpressEnv:
         self.addStepToHistory(infos)
 
         return observation, reward, isCompleted, infos
+
+    def serialize(self):
+        state = {
+            "gameName": self.gameName,
+            "gameParams": self.gameParams,
+            "seed": self.seed,
+            "gameFold": self.gameFold,
+            "envStepLimit": self.envStepLimit,
+            "generateGoldPath": self.generateGoldPath,
+            "actions": [info["lastActionStr"] for info in self.runHistory[1:]],
+        }
+        return state
+
+    @classmethod
+    def deserialize(cls, state):
+        env = cls(envStepLimit=state["envStepLimit"])
+        env.reset(
+            seed=state["seed"],
+            gameFold=state["gameFold"],
+            gameName=state["gameName"],
+            gameParams=state["gameParams"],
+            generateGoldPath=state["generateGoldPath"]
+        )
+        for action in state["actions"]:
+            env.step(action)
+
+        return env
+
+    def clone(self):
+        return self.deserialize(self.serialize())
